import aiohttp
from aiohttp import ClientTimeout

from nhm_spider.http.response import Response
from nhm_spider.common.log import get_logger


class Downloader:
    def __init__(self, spider):
        self.logger = get_logger(self.__class__.__name__)
        self.session = None
        self.spider = spider
        self.__headers = None
        self.__timeout = None
        self.__clear_cookie = None
        self.__use_session = None

    async def init(self):
        async def on_request_start(session, trace_config_ctx, params):
            # print("Starting request")
            pass

        async def on_request_end(session, trace_config_ctx, params):
            # print("Ending request")
            pass

        self.__headers = self.spider.settings.get("DEFAULT_REQUEST_HEADER", {})
        request_timeout = self.spider.settings.get("REQUEST_TIMEOUT", 180)
        self.__timeout = ClientTimeout(total=request_timeout)
        self.__clear_cookie = self.spider.settings.get("CLEAR_COOKIE", False)
        self.__use_session = self.spider.settings.get("USE_SESSION", True)

        trace_config = aiohttp.TraceConfig()
        trace_config.on_request_start.append(on_request_start)
        trace_config.on_request_end.append(on_request_end)

        self.session = aiohttp.ClientSession(headers=self.__headers, timeout=self.__timeout,
                                             trace_configs=[trace_config])

    async def send_request(self, request):
        try:
            # 每次请求前清除session缓存的cookies 为response set-cookie中自动缓存的
            if self.__clear_cookie is True:
                self.session.cookie_jar.clear()
            # 是否每次创建新session请求
            if self.__use_session is False:
                session = aiohttp.ClientSession(headers=self.__headers, timeout=self.__timeout)
                response = await self.send(session, request)
                await session.close()
            else:
                response = await self.send(self.session, request)
            if response is None:
                return
            # 获取完text之后，会自动关闭response。
            text = await response.text()  # TimeoutError
        except Exception as exception:
            return exception
        my_response = Response(request.url, request, text, response, response.status, response.headers)
        return my_response

    async def send(self, session, request):
        """ 处理不同method的请求参数 """
        if request.method.lower() == "get":
            response = await session.get(request.url, data=request.body, headers=request.headers,
                                         cookies=request.cookies, proxy=request.proxy)
        elif request.method.lower() == "post":
            response = await session.post(request.url, data=request.form, headers=request.headers,
                                          cookies=request.cookies, proxy=request.proxy)
        else:
            self.logger.error("传入不支持的方法。")
            response = None
        return response



from nhm_spider.download_middleware.default_headers import DefaultRequestHeadersDownloadMiddleware
from nhm_spider.download_middleware.retry import RetryDownloadMiddleware
from nhm_spider.download_middleware.timeout import TimeoutDownloadMiddleware

VERSION = 1.26
# 请求队列的最大值，控制此值可控制内存使用
REQUEST_QUEUE_MAX_SIZE = 0
# 是否使用session
USE_SESSION = True
# 是否清理session的cookie
# USE_SESSION = False时不生效
CLEAR_COOKIE = False
# 并发数量，即启动的任务数量
CONCURRENT_REQUESTS = 8
# 默认请求头
DEFAULT_REQUEST_HEADER = {
    'User-Agent': f'nhm-spider/{VERSION}'
}
# 是否开启调试日志
DEBUG = True
# 日志输出等级
DEBUG_LEVEL = "INFO"
# 默认请求超时时间，30秒
REQUEST_TIMEOUT = 30
# 默认开启的管道
ENABLED_PIPELINE = [
    # TmPipeline
    # TmQueryPipeline
]
# 默认开启的中间件
ENABLED_DOWNLOAD_MIDDLEWARE = [
    DefaultRequestHeadersDownloadMiddleware,
    RetryDownloadMiddleware,
    TimeoutDownloadMiddleware,
]
# 忽略的状态码错误
IGNORE_HTTP_ERROR = []



from nhm_spider.http.request import Request
from nhm_spider.common.log import get_logger
from nhm_spider.utils.project import get_default_settings


class Spider:
    start_urls = []
    settings = None
    custom_settings = {}

    def __init__(self, *args, **kwargs):
        self.logger = get_logger(self.__class__.__name__)
        self.logger.info(f"{self.__class__.__name__} start.")

    @classmethod
    def from_crawler(cls, crawler=None, *args, **kwargs):
        # todo: crawler is None
        spider = cls(*args, **kwargs)
        spider._set_crawler(crawler)
        spider._set_spider(crawler)
        return spider

    def _set_crawler(self, crawler):
        # crawler.spider = self
        pass

    def _set_spider(self, crawler):
        self.crawler = crawler
        # 获取 default_settings
        default_settings = get_default_settings()
        self.settings = default_settings | self.custom_settings
        self.DEBUG = self.settings.get("DEBUG", True)

    async def custom_init(self):
        pass

    def start_request(self):
        for url in self.start_urls:
            request = Request(url, callback=self.parse)
            yield request

    def parse(self, response):
        pass

    def __del__(self):
        self.logger.info(f"{self.__class__.__name__} closed.")



import json

from scrapy import Selector


class Response:
    def __init__(self, url=None, request=None, text=None, response=None, status=None, headers=None):
        self.url = url
        self.request = request
        self.text = text
        self.__response = response
        self.__selector = Selector(text=text)
        self.meta = request.meta
        self.status = status
        self.headers = headers

    def xpath(self, xpath_string):
        return self.__selector.xpath(xpath_string)

    def json(self):
        return json.loads(self.text)

    def __str__(self):
        return f"<{self.request.method.upper()} {self.url}>"



import aiomysql

from nhm_spider.common.log import get_logger


class DeprecateTmPipeline:
    pool: aiomysql.Pool

    def __init__(self):
        self.logger = get_logger(self.__class__.__name__)
        self.mysql_param = {
            "host": "127.0.0.1",
            "port": 3306,
            "db": "",
            "user": "",
            "password": ""
        }
        self.count = 0

    async def open_spider(self, spider):
        max_size = spider.settings.get("CONCURRENT_REQUESTS")
        self.pool = await aiomysql.create_pool(maxsize=max_size, **self.mysql_param)
        self.logger.info(f"Connect to mysql server {self.pool}.")

    async def process_item(self, item, spider):
        async with self.pool.acquire() as conn:
            async with conn.cursor(aiomysql.cursors.DictCursor) as cursor:
                # self.logger.info(pformat(item))
                self.count += 1
        return item

    async def close_spider(self, spider):
        self.pool.close()
        await self.pool.wait_closed()
        self.logger.info(f"Crawled total {self.count} items.")



import asyncio
import time

from nhm_spider.common.log import get_logger

logger = get_logger("TimeCounter")


def time_limit(seconds: (int, float) = 0, display=False):
    """
    装饰器
    统计使用时间
    :param display: 是否打印日志
    :param seconds: 控制方法执行的最短时间，在此时间之前完成，强行等待
    """
    def outer(func):
        def wrap(*args, **kwargs):
            start_time = time.time()
            r = func(*args, **kwargs)
            end_time = time.time()
            cost_time = end_time - start_time
            if display is True:
                logger.info(f"[{func.__name__}] cost time {cost_time}")
            if seconds and cost_time < seconds:
                time.sleep(seconds - cost_time)
            return r
        return wrap
    return outer


def async_time_limit(seconds: (int, float) = 0, display=False, timeout=None):
    """
    装饰器
    统计使用时间
    :param timeout: 协程对象超时时间
    :param display: 是否打印日志
    :param seconds: 控制方法执行的最短时间，在此时间之前完成，强行等待
    """
    def outer(func):
        async def wrap(*args, **kwargs):
            start_time = time.time()
            if timeout is None:
                r = await func(*args, **kwargs)
            else:
                r = await asyncio.wait_for(func(*args, **kwargs), timeout=timeout)
            end_time = time.time()
            cost_time = end_time - start_time
            if display is True:
                logger.info(f"[{func.__name__}] cost time {cost_time}")
            if seconds and cost_time < seconds:
                await asyncio.sleep(seconds - cost_time)
            return r
        return wrap
    return outer



from nhm_spider.download_middleware.default_headers import DefaultRequestHeadersDownloadMiddleware
from nhm_spider.download_middleware.retry import RetryDownloadMiddleware
from nhm_spider.download_middleware.timeout import TimeoutDownloadMiddleware
from nhm_spider import Item, Spider, Engine, Request, Field


class MpItem(Item):
    page = Field()


class MpSpider(Spider):
    custom_settings = {
        "USE_SESSION": True,
        "CLEAR_COOKIE": False,
        "CONCURRENT_REQUESTS": 4,
        "DEFAULT_REQUEST_HEADER": {
            'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) '
                          'Chrome/83.0.4103.97 Safari/537.36',
        },
        "ENABLED_PIPELINE": [
            # TmPipeline,
            # TmQueryPipeline,
        ],
    }

    def __init__(self):
        super(MpSpider, self).__init__()
        self.start_url = "http://www.mp.cc/search"
        self.page_url = "http://www.mp.cc/search/{}"

    async def start_request(self):
        request = Request(self.start_url, self.parse,
                          # proxy="http://127.0.0.1:8888"
                          )
        yield request

    def parse(self, response):
        page_info = response.xpath('//a[@class="number"][last()]/text()').get("0")
        total_page = int(page_info)
        for page in range(1, total_page + 1):
            request = Request(self.page_url.format(page), self.parse_page)
            request.meta["page"] = page
            yield request

    def parse_page(self, response):
        yield MpItem({"page": response.meta["page"]})


if __name__ == '__main__':
    engine = Engine()
    engine.run(MpSpider)



class Request:
    def __init__(self, url=None, callback=None, errback=None, form=None, headers=None, cookies=None,
                 method="get", meta=None, dont_filter=False, priority=None, proxy=None, *args, **kwargs):
        self.method: str = method
        self.url: str = url
        self.callback: callable = callback
        self.errback: callable = errback
        self.headers: dict = headers
        self.cookies: dict = cookies
        self.proxy: dict = proxy
        self.form: dict = form
        self.meta: dict = meta or {}
        self.dont_filter: bool = dont_filter
        # 优先级，越小越先请求
        self.priority: int = priority if priority is None else -priority

        # 创建body，供指纹使用
        if method.lower() == "get":
            self.body = url.split("?", 1)[-1] if "?" in url else ""
        elif method.lower() == "post":
            self.body = "&".join(map(lambda _: "=".join(_), sorted(form.items(), key=lambda x: x[0])))
        else:
            self.body = ""
        self.body = self.body.encode()

    # 供优先级队列比较，比较的是Request.priority的大小

    def __lt__(self, other):
        return self.priority < other.priority

    def __gt__(self, other):
        return self.priority > other.priority

    def __str__(self):
        return f"<{self.method.upper()} {self.url}>"


class FormRequest(Request):
    def __init__(self, url=None, callback=None, errback=None, form=None, headers=None, cookies=None, meta=None,
                 dont_filter=False, priority=None, proxy=None, *args, **kwargs):
        self.method = "post"
        super(FormRequest, self).__init__(url, callback, errback, form, headers, cookies, self.method, meta,
                                          dont_filter, priority, proxy, *args, **kwargs)



import asyncio
from traceback import format_exc
from types import GeneratorType, AsyncGeneratorType
from inspect import isawaitable, iscoroutine

from scrapy.utils.request import request_fingerprint

from nhm_spider.common.log import get_logger
from nhm_spider.utils.pqueue import SpiderPriorityQueue
from nhm_spider.utils.signal import SignalManager
from nhm_spider.item import Item
from nhm_spider.http import Request, Response


class Scheduler:
    def __init__(self, spider):
        self.logger = get_logger(self.__class__.__name__)
        self.request_queue = None
        self.signal_manager = None
        # 请求去重队列
        self.dupe_memory_queue = set()
        self.spider = spider
        self.item_count = 0
        self.request_count = 0
        self.tasks = []

        settings = spider.settings

        # pipeline
        self.concurrent_requests = settings.get("CONCURRENT_REQUESTS", 8)
        enabled_pipeline = settings.get("ENABLED_PIPELINE", [])
        self.enabled_pipeline = [cls() for cls in enabled_pipeline]
        # download middleware
        enabled_download_middleware = settings.get("ENABLED_DOWNLOAD_MIDDLEWARE", [])
        self.enabled_download_middleware = [cls() for cls in enabled_download_middleware]
        # spider middleware
        # enabled_spider_middleware = settings.get("ENABLED_SPIDER_MIDDLEWARE", [])
        # self.enabled_spider_middleware = [cls() for cls in enabled_spider_middleware]

    async def init(self):
        self.request_queue = SpiderPriorityQueue()
        self.signal_manager = SignalManager(self.request_queue)
        self.signal_manager.connect()

    async def next_request(self):
        request = await self.request_queue.get()
        return request

    async def crawl(self, spider, downloader):
        """
        协程主程序
        """
        await self.init()
        await downloader.init()
        await spider.custom_init()

        # init pipeline
        for pipeline in self.enabled_pipeline:
            # 确认是否使用的异步的pipeline
            if not hasattr(pipeline, "open_spider"):
                continue
            pip = pipeline.open_spider(spider)
            if isawaitable(pip):
                await pip

        # init download middleware
        for middleware in self.enabled_download_middleware:
            # 确认是否使用的异步的middleware
            if not hasattr(middleware, "open_spider"):
                continue
            mid = middleware.open_spider(spider)
            if isawaitable(mid):
                await mid

        tasks = self.tasks
        try:
            # 初始化
            results = spider.start_request()
            await self.process_results(results)
            for task_number in range(self.concurrent_requests):
                task = asyncio.create_task(self.process(downloader))
                tasks.append(task)

            # 阻塞并等待所有任务完成
            tasks.append(asyncio.create_task(self.heartbeat()))
            await self.request_queue.join()

        finally:
            # clear pipeline
            for pipeline in self.enabled_pipeline:
                # 确认是否使用的异步的pipeline
                if not hasattr(pipeline, "close_spider"):
                    continue
                pip = pipeline.close_spider(spider)
                if isawaitable(pip):
                    await pip

            # clear download middleware
            for middleware in self.enabled_download_middleware:
                # 确认是否使用的异步的middleware
                if not hasattr(middleware, "close_spider"):
                    continue
                mid = middleware.close_spider(spider)
                if isawaitable(mid):
                    await mid

            await downloader.session.close()
            # 所有task完成后，取消任务，退出程序
            for task in tasks:
                task.cancel()
            # 等待task取消完成
            await asyncio.gather(*tasks, return_exceptions=True)

    async def heartbeat(self, heartbeat_interval=60):
        """
        todo: 统计采集的页数，抓取的条数。
              考虑放到extensions中去
        """
        last_item_count = 0
        last_request_count = 0
        while True:
            request_speed = self.request_count - last_request_count
            last_request_count = self.request_count
            item_speed = self.item_count - last_item_count
            last_item_count = self.item_count
            self.logger.info(f"Crawled {last_request_count} pages (at {request_speed} pages/min), "
                             f"scraped {last_item_count} items (at {item_speed} items/min), "
                             f"queue size {self.request_queue.qsize()}.")

            await asyncio.sleep(heartbeat_interval)

    async def process_results(self, results, response=None):
        if results:
            if isinstance(results, GeneratorType):
                for obj in results:
                    await self.process_result_single(obj, response)
            elif isinstance(results, AsyncGeneratorType):
                async for obj in results:
                    await self.process_result_single(obj, response)
            elif isinstance(results, Request):
                await self.process_result_single(results, response)
            elif iscoroutine(results):
                await results
            else:
                # todo: 考虑如何处理
                self.logger.error(f"丢弃该任务，未处理的处理结果类型：{results}。")

    async def process_result_single(self, obj, response):
        if isinstance(obj, Request):
            # 处理request对象优先级，深度优先
            if obj.priority is None:
                if response is not None:
                    obj.priority = response.request.priority - 1
                else:
                    obj.priority = 0

            fp = request_fingerprint(obj)
            # 根据指纹去重。
            if obj.dont_filter is True or fp not in self.dupe_memory_queue:
                obj.fp = fp
                await self.enqueue_request(obj)
                self.dupe_memory_queue.add(fp)

        elif isinstance(obj, Item):
            if not self.enabled_pipeline and self.spider.DEBUG is True:
                self.logger.info(obj)
            self.item_count += 1

            for pipeline in self.enabled_pipeline:
                # 确认是否使用的异步的pipeline
                if not hasattr(pipeline, "process_item"):
                    continue
                obj = pipeline.process_item(obj, self.spider)
                if isawaitable(obj):
                    obj = await obj

        else:
            self.logger.warning(f"[yield]尚未处理的类型[{obj.__class__.__name__}]。")

    async def enqueue_request(self, request):
        assert request.callback, "未指定回调方法。"
        await self.request_queue.put(request)

    async def process(self, downloader):
        while True:
            request = await self.next_request()

            response = await self.download_request(request, downloader)
            if not isinstance(response, Response):
                # todo: 待处理非response的情况

                # 失败的请求也要调用task_done，否则无法结束。
                self.request_queue.task_done()
                self.request_count += 1
                continue
            else:
                if self.spider.DEBUG is True:
                    self.logger.info(f"Crawled ({response.status}) {response}.")

            # todo: process_spider_in
            results = request.callback(response)
            # todo: process_spider_out 非此位置
            try:
                await self.process_results(results, response)
            except:
                self.logger.error(format_exc())
            finally:
                self.request_queue.task_done()
                self.request_count += 1

    async def download_request(self, request, downloader):
        # process_request
        for middleware in self.enabled_download_middleware:
            # 确认是否使用的异步的middleware
            if not hasattr(middleware, "process_request"):
                continue
            result = middleware.process_request(request, self.spider)
            if isawaitable(result):
                result = await result

            if result is None:
                pass
            elif isinstance(result, Request):
                return await self.process_results(result)
            elif isinstance(result, Response):
                # 返回response则直接跳过process_request
                request = result
                break
            else:
                self.logger.error(f"未知的对象类型，{request}。")
                raise TypeError("未知的对象类型")

        if isinstance(request, Request):
            response = await downloader.send_request(request)
        elif isinstance(request, Response):
            response = request
        else:
            self.logger.error(f"未知的对象类型，{request}。")
            raise TypeError("未知的对象类型")

        # process_response
        if isinstance(response, Response):
            for middleware in self.enabled_download_middleware:
                # 确认是否使用的异步的middleware
                if not hasattr(middleware, "process_response"):
                    continue
                result = middleware.process_response(request, response, self.spider)
                if isawaitable(result):
                    result = await result

                if result is None:
                    pass
                elif isinstance(result, Request):
                    return await self.process_results(result)
                elif isinstance(result, Response):
                    response = result
                    break
        elif isinstance(response, Exception):
            for middleware in self.enabled_download_middleware:
                # 确认是否使用的异步的middleware
                if not hasattr(middleware, "process_exception"):
                    continue
                result = middleware.process_exception(request, response, self.spider)
                if isawaitable(result):
                    result = await result

                if result is None:
                    pass
                elif isinstance(result, Request):
                    return await self.process_results(result)
                elif isinstance(result, Response):
                    response = result
                    break

        else:
            self.logger.error(f"未知的Response类型，{response}。")
            raise TypeError("未知的Response类型")

        return response



from pprint import pformat


# todo: 待增加不同类型的字段，增加字段`类型检查`或`自动转换`功能。
#       IntegerField, StringField, FloatField, JsonField ...
#       再深入可考虑`长度检查`等
class Field:
    pass


class Item:
    __meta = set()

    def __init__(self, fields: dict = None, **kwargs):
        self.__values = dict()
        for attr in dir(self):
            if not attr.startswith("_") and isinstance(getattr(self, attr), Field):
                self.__meta.add(attr)
                self.__values[attr] = None

        if fields:
            for field in fields:
                if field not in self.__meta:
                    raise AttributeError(f"Class {self.__class__.__name__} not exists field [{field}].")
                self.__values[field] = fields[field]

        for field in kwargs:
            if field not in self.__meta:
                raise AttributeError(f"Class {self.__class__.__name__} not exists field [{field}].")
            self.__values[field] = kwargs[field]

    def __setitem__(self, key, value):
        if isinstance(self.__getattribute__(key), Field):
            self.__values[key] = value

    def __getitem__(self, item):
        if item in self.__meta:
            return self.__values[item]

    def __delitem__(self, key):
        del self.__values[key]

    def __setattr__(self, key, value):
        if key in self.__meta:
            raise AttributeError(f"Class {self.__class__.__name__} field {key} not enable rewrite.")
        super().__setattr__(key, value)

    def __iter__(self):
        return iter(self.__values)

    def __repr__(self):
        return f"{self.__class__.__name__}({pformat(self.__values)})"



class DefaultRequestHeadersDownloadMiddleware:
    def __init__(self):
        self.default_headers = None

    def open_spider(self, spider):
        self.default_headers = spider.settings.get("DEFAULT_REQUEST_HEADER", {})

    def process_request(self, request, spider):
        # todo: 处理headers大小写为统一规则，否则合并会出现多个相同headers。
        if request.headers is None:
            request.headers = self.default_headers
        else:
            # 合并默认headers和当前request的headers
            request.headers = self.default_headers | request.headers
        return None



from asyncio.exceptions import TimeoutError

from nhm_spider.common.log import get_logger


class TimeoutDownloadMiddleware:
    def __init__(self):
        self.logger = get_logger(self.__class__.__name__)

        # 最大重试次数
        self.max_retry_times = 3

    def process_exception(self, request, exception, spider):
        if isinstance(exception, TimeoutError):
            request.dont_filter = True
            retry_times = request.meta.get("exception_retry_times", 0)
            if retry_times < self.max_retry_times:
                self.logger.info(f"{request} exception: TimeoutError, retry {retry_times + 1} time...")
                request.meta["exception_retry_times"] = retry_times + 1
                return request
            else:
                self.logger.warning(f"{request} exception error, retry {self.max_retry_times} times error.")
                return None
        return exception



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path
# nhm custom
from nhm_spider.settings.default_settings import VERSION

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='nhm_spider',
    version=str(VERSION),
    author='noHairMan',
    author_email='zongxuheng@163.com',
    url="https://github.com/noHairMan/nhm-spider",
    description='base on asyncio, spider module like scrapy.',
    long_description=long_description,  # 这里是文档内容, 读取readme文件
    long_description_content_type='text/markdown',  # 文档格式
    packages=find_packages(),
    classifiers=[  # 这里我们指定证书, python版本和系统类型
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',  # 这里指定python版本号必须大于3.9才可以安装
    install_requires=["aiomysql", "scrapy", "aiohttp"]  # 我们的模块所用到的依赖, 这里指定的话, 用户安装你的模块时, 会自动安装这些依赖
)



import asyncio

from nhm_spider.common.log import get_logger
from nhm_spider.common.time_counter import time_limit
from nhm_spider.core.downloader import Downloader
from nhm_spider.core.scheduler import Scheduler


class Engine:
    def __init__(self):
        self.logger = get_logger(self.__class__.__name__)

    # 显示方法执行的时间
    @time_limit(display=True)
    def run(self, spider_class):
        spider = spider_class.from_crawler()
        if not hasattr(spider, "custom_settings"):
            spider.custom_settings = {}

        downloader = Downloader(spider)
        scheduler = Scheduler(spider)

        asyncio.run(scheduler.crawl(spider, downloader))

    def __del__(self):
        self.logger.info("Engine quit.")



from logging import Formatter, getLogger, StreamHandler, INFO, Logger, FileHandler, CRITICAL


__instance = {}


def get_logger(name="common", log_level=None, formatter=None):
    """
    单角色模式
    """
    if name in __instance:
        return __instance[name]
    formatter_option = formatter or f'%(asctime)s [%(name)s] %(levelname)s: %(message)s'
    logger_formatter = Formatter(formatter_option)
    handler = StreamHandler()
    handler.setFormatter(logger_formatter)
    handler.setLevel(log_level or INFO)
    _logger = getLogger(name)
    _logger.addHandler(handler)
    _logger.setLevel(INFO)
    __instance[name] = _logger
    return _logger


def add_format(_logger: Logger, formatter):
    logger_formatter = Formatter(formatter)
    for handler in _logger.handlers:
        handler.setFormatter(logger_formatter)


def add_file_handler(_logger: Logger, file_name, log_level=None):
    """
    使用文件存储日志。
    """
    handler = FileHandler(file_name)
    for ori_handler in _logger.handlers:
        if isinstance(ori_handler, StreamHandler):
            handler.setFormatter(ori_handler.formatter)
    handler.setLevel(log_level or INFO)
    _logger.addHandler(handler)


def get_file_logger(name=None, file_name=None, stream_log_level=None, file_log_level=None):
    assert file_name, "日志文件名不能为空。"
    _logger = get_logger(name, stream_log_level)
    add_file_handler(_logger, file_name, file_log_level)
    return _logger


def test():
    file_name = "/disk1t/logs/test.log"
    logger = get_file_logger(file_name=file_name, stream_log_level=CRITICAL)
    formatter_option = f'%(asctime)s [%(name)s] %(levelname)s: %(message)s  测试!'
    add_format(logger, formatter_option)
    logger.info("123")


if __name__ == '__main__':
    test()




import signal

from nhm_spider.common.log import get_logger


class SignalManager:
    def __init__(self, request_queue):
        self.logger = get_logger(self.__class__.__name__)
        self.request_queue = request_queue

    def handler(self, signum, frame):
        # 执行后会释放 join 部分
        self.logger.warning(f"收到信号[{signum}]。")
        self.request_queue.close()

    def connect(self):
        signal.signal(signal.SIGINT, self.handler)

    def __del__(self):
        pass



from nhm_spider.http.response import Response
from nhm_spider.common.log import get_logger


class RetryDownloadMiddleware:
    def __init__(self):
        self.logger = get_logger(self.__class__.__name__)
        # 最大重试次数
        self.max_retry_times = 3
        self.ignore_http_error = None

    def open_spider(self, spider):
        self.ignore_http_error = spider.settings.get("IGNORE_HTTP_ERROR", [])

    def process_response(self, request, response, spider):
        if request.meta.get('dont_retry', False):
            return response

        if isinstance(response, Response) and response.status != 200:
            if response.status in self.ignore_http_error:
                # 处理忽略的错误状态码
                return response

            if request.dont_filter is not True:
                request.dont_filter = True
            retry_times = request.meta.get("retry_times", 0)
            if retry_times < self.max_retry_times:
                self.logger.info(f"({response.status}) {response} status error, retry {retry_times + 1} time...")
                request.meta["retry_times"] = retry_times + 1
                return request
            else:
                self.logger.warning(f"{response} retry max {self.max_retry_times} times error。")
                return None
        return response



import time
from datetime import datetime


def timestamp_to_date(timestamp, format_string="%Y-%m-%d %H:%M:%S"):
    """
    时间戳转为日期
    """
    return time.strftime(format_string, time.localtime(timestamp))


def date_to_timestamp(date_string, format_string="%Y-%m-%d %H:%M:%S"):
    """
    日期转换为时间戳
    """
    return int(time.mktime(time.strptime(date_string, format_string)))


def get_date(format_string='%Y-%m-%d'):
    return datetime.now().strftime(format_string)
